#include <iostream> // pour utiliser std::cout std::endl << >>
#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>
#include <glm/gtx/transform.hpp> // POUR UTILISER glm::ortho()
#include <glm/gtc/type_ptr.hpp> // POUR UTILISER glm::value_ptr
#include <glm/gtc/matrix_inverse.hpp> // POUR UTILISER glm::inverse()

#include "main.h"
#include "LisEtCompileLesShaders.h"


#define TAILLE_HORIZONTALE 1400
#define TAILLE_VERTICALE 800

GLFWwindow *MaFenetre;
glm::mat4 MaMatriceDeProjection;

//----------------------------------------
// CLES SUR LA CARTE GRAPHIQUE LIEES AU PROGRAMME GPU DE TESSELLATION

// CLE DU PROGRAMME GPU
static GLuint MonProgrammeGLSL;

// CLE DES VARIABLES UNIFORMES DANS LE PROGRAMME GLSL
static GLint MonSommetID;
static GLint FinesseDecoupageID;
static GLint MaMatriceDeProjectionID;
static GLint MonTempsID;
static GLint MaForceTotaleID;

//----------------------------------------
// CLES DE LA GEOMETRIE ET DES INDICES DES SOMMETS DES PATCHS SUR LA CARTE GRAPHIQUE
static GLuint MonVAO;
static GLuint MesSommetsID;
static GLuint MesIndicesID;

glm::vec3 MesSommets[]= {// 4 SOMMETS POUR DEFINIR UNE COURBE DE BEZIER DE DEGRE 3
	glm::vec3(2.0, 0.0, 0.),
	glm::vec3(-1.0, 0.0, 0.),
	glm::vec3(0.0,  0.0, 0.),
	glm::vec3( 1.0, 0.0, 0.),
}; 
static glm::vec3 MesVitesses[4]={
	glm::vec3(0.0, 0.0, 0.0),
	glm::vec3(0.0, 0.0, 0.0),
	glm::vec3(0.0, 0.0, 0.0),
	glm::vec3(0.0, 0.0, 0.0)};
static glm::vec3 MesForces[4]={
	glm::vec3(0.0, 0.0, 0.0),
	glm::vec3(0.0, 0.0, 0.0),
	glm::vec3(0.0, 0.0, 0.0),
	glm::vec3(0.0, 0.0, 0.0)};

void InitialisationDuProgrammeGPUEtDesLiensVersCeProgramme()
{
	MonProgrammeGLSL = LisEtCompileLesShaders( 
		"../shader/tessellation.vert.glsl",
		"../shader/tessellation.cont.glsl",
		"../shader/tessellation.eval.glsl",
		"../shader/tessellation.geom.glsl",
		"../shader/tessellation.frag.glsl");

// JE RECHERCHE LES CLES DESIGNANT LES VARIABLES UNIFORMES QUE J UTILISE DANS LES CODES GLSL
// CES VARIABLES SONT DITES UNIFORMES CAR ELLES ONT LA MEME VALEUR POUR TOUS LES SOMMETS/PRIMITIVES/FRAGMENTS.
// LA CHAINE DE CARACTERE A DROITE CONTIENT LE NOM DES VARIABLES DANS LES CODES GLSL
// LE RESULTAT DE LA FONCTION EST LA CLE VERS CETTE VARIABLE GLSL
// CETTE CLE EST UTILISEE PLUS TARD DANS LE CODE C++ POUR TRANSMETTRE UNE VALEUR DU CPU VERS LA VARIABLE GPU
	FinesseDecoupageID = glGetUniformLocation(MonProgrammeGLSL, "u_FinesseDecoupage");
	MonSommetID = glGetAttribLocation(MonProgrammeGLSL, "MonSommetEnEntree");
	MaMatriceDeProjectionID = glGetUniformLocation(MonProgrammeGLSL, "MaMatriceDeProjection");
	MaForceTotaleID = glGetUniformLocation(MonProgrammeGLSL, "MaForceTotale");
	MonTempsID = glGetUniformLocation(MonProgrammeGLSL, "MonTemps");
	fprintf( stderr, "MonSommetID = %d\n", MonSommetID);
	fprintf( stderr, "FinesseDecoupageID = %d\n", FinesseDecoupageID);
	fprintf( stderr, "MonTempsID = %d\n", MonTempsID);
}


void InitialisationDesDonnees()
{
	GLuint MesIndices[]= { 0, 1, 2, 3}; 

	glGenBuffers( 1, &MesSommetsID); // BUFFER DES SOMMETS
	glBindBuffer(GL_ARRAY_BUFFER, MesSommetsID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(MesSommets), (void*)MesSommets, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glGenBuffers( 1, &MesIndicesID); // BUFFER DES INDICES
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, MesIndicesID);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(MesIndices), (void*)MesIndices, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

}

glm::vec3 ForceRessort( float DistanceAuRepos, float KS, float KD, glm::vec3 P1, glm::vec3 V1, glm::vec3 P2, glm::vec3 V2)
{
	glm::vec3 l = P2 - P1;
	glm::vec3 dl = V2 - V1;
	GLfloat Distance = sqrt(glm::dot(l,l));
	GLfloat Coeff = KS * (Distance - DistanceAuRepos) + KD * glm::dot(dl, l)/Distance;
	glm::vec3 Force = Coeff / Distance * l;
	return Force;
}

glm::vec3 PositionSourisEspaceObjet()
{
	double CurseurX, CurseurY;
	int Largeur, Hauteur;

	glfwGetFramebufferSize(MaFenetre, &Largeur, &Hauteur);
	glm::vec4 MonViewPort = glm::vec4( 0, 0, Largeur, Hauteur);
//	std::cout << "Largeur :  " << Largeur << std::endl;
//	std::cout << "hauteur :  " << Hauteur << std::endl;

	glfwGetCursorPos( MaFenetre, &CurseurX, &CurseurY);
	glm::vec3 SommetFenetre = glm::vec3( CurseurX, Hauteur - CurseurY, -0.001);

	glm::vec3 SommetSouris = glm::unProject( SommetFenetre, glm::mat4(1.0), MaMatriceDeProjection, MonViewPort);
	SommetSouris.z = 0.;

	return SommetSouris;
}

void AfficheSommet( glm::vec3 S)
{
	std::cout << S[0] <<  " " << S[1] << " " << S[2] << std::endl;
}

void InitialisationDesSommetsEnFonctionDeLaSouris()
{
	glm::vec3 SommetSouris = PositionSourisEspaceObjet();
	glm::vec3 DirectionCourbe = glm::vec3( 1., 0., 0.);

	MesSommets[0] = glm::mix( SommetSouris+DirectionCourbe, SommetSouris, 0.99);
	MesSommets[1] = MesSommets[0]+DirectionCourbe;
	MesSommets[2] = glm::mix( MesSommets[0], MesSommets[1], 2.);
	MesSommets[3] = glm::mix( MesSommets[1], MesSommets[2], 2.);
}


glm::vec3 ForceDeLaSouris()
{


	glm::vec3 SommetSouris = PositionSourisEspaceObjet();

//	std::cout << "SommetSouris = " << SommetSouris.x << " " << SommetSouris.y << " " << SommetSouris.z << std::endl;
	glm::vec3 VitesseSouris = glm::vec3( 0., 0., 0.);
	glm::vec3 Force = ForceRessort( 0.01, 100., 90., SommetSouris, VitesseSouris, MesSommets[0], MesVitesses[0]);

	return Force;
}

void AnimationDesSommets()
{
	static GLfloat MonTemps = 0.f;
	static GLfloat MaForceTotale;

// TRANSFERT DU COMPTEUR TEMPS AU PROGRAMME GLSL
	MonTemps += 0.1;
	glUniform1f( MonTempsID, MonTemps);

// MISE A ZERO DES FORCES
	for (int i=0; i<4; i++) {
		MesForces[i] = glm::vec3(0.f, 0.f, 0.f);
	}

// CALCUL DES FORCES MASSE RESSORT
/*
*/
	for (int i=0; i<3; i++) {
		glm::vec3 Force = ForceRessort(  1., 4., 50., MesSommets[i], MesVitesses[i], MesSommets[i+1], MesVitesses[i+1]);
		MesForces[i] = MesForces[i] + Force;
		MesForces[i+1] = MesForces[i+1] - Force;
	}
	for (int i=0; i<2; i++) {
		glm::vec3 Force = ForceRessort(  2., 4., 20., MesSommets[i], MesVitesses[i], MesSommets[i+2], MesVitesses[i+2]);
		MesForces[i] = MesForces[i] + Force;
		MesForces[i+2] = MesForces[i+2] - Force;
	}

// FORCE DE LA SOURIS
/*
*/
	MesForces[0] = MesForces[0] - ForceDeLaSouris();


// FORCE GLOBALE DE DECELLERATION
/*
*/
	for (int i=0; i<4; i++) {
		MesForces[i] = MesForces[i] - glm::mix( glm::vec3(0.), MesVitesses[i], 0.5);
	}


// INTEGRATION DES FORCES
	MaForceTotale = 0.0;
	for (int i=0; i<4; i++) {
		MaForceTotale += glm::dot( MesForces[i], MesForces[i]);
		const GLfloat DeltaT = 0.01;
		MesVitesses[i] = MesVitesses[i] + DeltaT * MesForces[i];
		MesSommets[i] = MesSommets[i] + DeltaT * MesVitesses[i];
	}
	MaForceTotale = sqrt(MaForceTotale);

// TRANSFERT DE LA FORCE TOTALE AU PROGRAME GLSL
	glUniform1f( MaForceTotaleID, MaForceTotale);

// ECRITURE DES SOMMETS SUR LE GPU
	glBindBuffer(GL_ARRAY_BUFFER, MesSommetsID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(MesSommets), (void*)MesSommets, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

}

void EncapsuleLesDonneesDansUnVao()
{
	glGenVertexArrays( 1, &MonVAO); // ON GENERE UN VAO
	glBindVertexArray(MonVAO);

	glEnableVertexAttribArray(0); // ON ACTIVE LE TABLEAU D'ATTRIBUT D'INDICE 0

        glBindBuffer(GL_ARRAY_BUFFER, MesSommetsID); // ON ACTIVE LE VBO DES SOMMETS
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, MesIndicesID); // ON ACTIVE LE VBO DES INDICES

       // ON INDIQUE COMMENT LES DONNEES SONT TRANSMISES AU PROGRAMME
	glVertexAttribPointer(
            0,                  // indice de l'attribut à configurer, correspond à la ligne "layout(location=0) in vec3 MonSommetEnEntree;" dans le vertex shader
            3,                  // nombre de composante par position. ici 3 composantes: (x, y, z).
            GL_FLOAT,           // type des composantes
            GL_FALSE,           // normalisation des composantes
            0,                  // decalage des composantes
            (void*)0            // offset des composantes
        );
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);
}



int main( int argc, char *argv[])
{
	MaFenetre = InitialiseUnContexteEtUneFenetreOpenGL();

	InitialisationDuProgrammeGPUEtDesLiensVersCeProgramme();
	glUseProgram(MonProgrammeGLSL);

	InitialisationDesDonnees();
	EncapsuleLesDonneesDansUnVao();

	std::cout << "MonProgrammeGLSL = " << MonProgrammeGLSL << std::endl;

	std::cout<<"MonVAO = "<<MonVAO<< std::endl;

	glClearColor( 0.2, 0.2, 0.2, 1.0f);

	glClearDepth(1.0f);

	glEnable(GL_DEPTH_TEST);

// INDIQUE LE NOMBRE DE POINTS A TRAITER PAR PATCH
	glPatchParameteri(GL_PATCH_VERTICES, 4);

// LES LIGNES SUIVANTES PEUVENT PERMETTRE DE SE PASSER DE TESSELATION CONTROL SHADER DANS LES CAS SIMPLES
	GLfloat DefaultOuterLevel[]={4.,4.,4.,4.};
	GLfloat DefaultInnerLevel[]={4.,4.};
	glPatchParameterfv(GL_PATCH_DEFAULT_OUTER_LEVEL, DefaultOuterLevel);
	glPatchParameterfv(GL_PATCH_DEFAULT_INNER_LEVEL, DefaultInnerLevel);

	GLfloat Aspect = (1.0f * TAILLE_HORIZONTALE) / TAILLE_VERTICALE;
	MaMatriceDeProjection = glm::ortho(  -5.f*Aspect, 5.f*Aspect, -5.f, 5.f, 0.1f, 100.f);

	InitialisationDesSommetsEnFonctionDeLaSouris();
	
	while (!glfwWindowShouldClose( MaFenetre)) {

		AnimationDesSommets();

		TraiteLesEvenementsClavier( MaFenetre);

		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

		glUniform1f( FinesseDecoupageID, (GLfloat) (40.0f)); // FINESSE DE LA TESSELLATION
		glUniformMatrix4fv( MaMatriceDeProjectionID, 1, GL_FALSE, glm::value_ptr( MaMatriceDeProjection)); // FINESSE DE LA TESSELLATION

		glBindVertexArray( MonVAO);
		glDrawElements( GL_PATCHES, 4, GL_UNSIGNED_INT, 0); // DESSIN D'UNE COURBE DE BEZIER DE DEGRE 3
		glBindVertexArray( 0);

		glfwSwapBuffers(MaFenetre);

		glfwPollEvents();
	}

	glDeleteBuffers(1, &MesIndicesID);
	glDeleteBuffers(1, &MesSommetsID);
	glDeleteVertexArrays(1, &MonVAO);

	glfwTerminate();
}


GLFWwindow* InitialiseUnContexteEtUneFenetreOpenGL()
{
	GLFWwindow* MaFenetre;

	// INITIALISATION DE GLFW
	if( !glfwInit() )
	{
		std::cout << "Echec de l'initialisation de GLFW" << std::endl;
		exit(EXIT_FAILURE);
	}

	glfwWindowHint(GLFW_SAMPLES, 4); // Anti Aliasing
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3); // OpenGL 3.3
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);


	// Ouverture d'une fenêtre en 1024x768
	// et creation d'un contexte OpenGL
	if( !(MaFenetre = glfwCreateWindow(TAILLE_HORIZONTALE, TAILLE_VERTICALE, "Polytech_RICM4", NULL,NULL)) )
//	if( !(MaFenetre = glfwCreateWindow(2560, 1920, "Polytech_RICM4", glfwGetPrimaryMonitor(),NULL)) )
	{
		std::cout << "Echec de l'ouverture de fenetre OpenGL" << std::endl;
		glfwTerminate();
		exit(EXIT_FAILURE);
	}
	glfwMakeContextCurrent(MaFenetre);

	// Autorise GLFW a recevoir les appuis de touche
	glfwSetInputMode( MaFenetre, GLFW_STICKY_KEYS, GL_TRUE);

	// Initialisation de  GLEW
	glewExperimental = GL_TRUE; 
	if (glewInit() != GLEW_OK) {
		std::cout << "Echec de l'initialisation de GLEW" << std::endl;
		exit(EXIT_FAILURE);
	}

	// Affichage des donnees du contexte OpenGL
	const GLubyte* renderer = glGetString (GL_RENDERER);
	std::cout << "TYPE DE CARTE GRAPHIQUE : " << renderer << std::endl;

	const GLubyte* version = glGetString (GL_VERSION);
	std::cout << "VERSION DE PILOTE OpenGL : " << version << std::endl;

	const GLubyte* shading_version = glGetString (GL_SHADING_LANGUAGE_VERSION);
	std::cout << "VERSION GLSL: " << shading_version << std::endl;

	return MaFenetre;
}

void TraiteLesEvenementsClavier(GLFWwindow *MaFenetre)
{
	if (glfwGetKey( MaFenetre, GLFW_KEY_A ) == GLFW_PRESS) {
		std::cout << "Touche A\n";
	}
}


